package part1_test

import (
	"fmt"
	"log"
	"strings"
	"testing"

	"github.com/google/go-cmp/cmp"

	"source.toby3d.me/toby3d/adventofcode/internal/day2/part1"
)

func TestParseGame(t *testing.T) {
	t.Parallel()

	for name, tc := range map[string]struct {
		input  string
		expect part1.Game
	}{
		"1": {
			input: "Game 1: 3 blue, 4 red; 1 red, 2 green, 6 blue; 2 green",
			expect: part1.Game{ID: 1, Rounds: []part1.Round{
				{Blue: 3, Red: 4},
				{Red: 1, Green: 2, Blue: 6},
				{Green: 2},
			}},
		},
		"2": {
			input: "Game 2: 1 blue, 2 green; 3 green, 4 blue, 1 red; 1 green, 1 blue",
			expect: part1.Game{ID: 2, Rounds: []part1.Round{
				{Blue: 1, Green: 2},
				{Green: 3, Blue: 4, Red: 1},
				{Green: 1, Blue: 1},
			}},
		},
		"3": {
			input: "Game 3: 8 green, 6 blue, 20 red; 5 blue, 4 red, 13 green; 5 green, 1 red",
			expect: part1.Game{ID: 3, Rounds: []part1.Round{
				{Green: 8, Blue: 6, Red: 20},
				{Blue: 5, Red: 4, Green: 13},
				{Green: 5, Red: 1},
			}},
		},
		"4": {
			input: "Game 4: 1 green, 3 red, 6 blue; 3 green, 6 red; 3 green, 15 blue, 14 red",
			expect: part1.Game{ID: 4, Rounds: []part1.Round{
				{Green: 1, Red: 3, Blue: 6},
				{Green: 3, Red: 6},
				{Green: 3, Blue: 15, Red: 14},
			}},
		},
		"5": {
			input: "Game 5: 6 red, 1 blue, 3 green; 2 blue, 1 red, 2 green",
			expect: part1.Game{ID: 5, Rounds: []part1.Round{
				{Red: 6, Blue: 1, Green: 3},
				{Blue: 2, Red: 1, Green: 2},
			}},
		},
	} {
		name, tc := name, tc

		t.Run(name, func(t *testing.T) {
			t.Parallel()

			actual, err := part1.ParseGame(tc.input)
			if err != nil {
				t.Fatal(err)
			}

			if diff := cmp.Diff(*actual, tc.expect); diff != "" {
				t.Error(diff)
			}
		})
	}
}

func TestGame_IsPossible(t *testing.T) {
	t.Parallel()

	bag := part1.Bag{
		Red:   12,
		Green: 13,
		Blue:  14,
	}

	for name, tc := range map[string]struct {
		input  part1.Game
		expect bool
	}{
		"1": {
			input: part1.Game{ID: 1, Rounds: []part1.Round{
				{Blue: 3, Red: 4},
				{Red: 1, Green: 2, Blue: 6},
				{Green: 2},
			}},
			expect: true,
		},
		"2": {
			input: part1.Game{ID: 2, Rounds: []part1.Round{
				{Blue: 1, Green: 2},
				{Green: 3, Blue: 4, Red: 1},
				{Green: 1, Blue: 1},
			}},
			expect: true,
		},
		"3": {
			input: part1.Game{ID: 3, Rounds: []part1.Round{
				{Green: 8, Blue: 6, Red: 20},
				{Blue: 5, Red: 4, Green: 13},
				{Green: 5, Red: 1},
			}},
			expect: false,
		},
		"4": {
			input: part1.Game{ID: 4, Rounds: []part1.Round{
				{Green: 1, Red: 3, Blue: 6},
				{Green: 3, Red: 6},
				{Green: 3, Blue: 15, Red: 14},
			}},
			expect: false,
		},
		"5": {
			input: part1.Game{ID: 5, Rounds: []part1.Round{
				{Red: 6, Blue: 1, Green: 3},
				{Blue: 2, Red: 1, Green: 2},
			}},
			expect: true,
		},
	} {
		name, tc := name, tc

		t.Run(name, func(t *testing.T) {
			t.Parallel()

			if actual := tc.input.IsPossible(bag); actual != tc.expect {
				t.Errorf("want %t, got %t", tc.expect, actual)
			}
		})
	}
}

func Example() {
	// You play several games and record the information from each game
	// (your puzzle input). Each game is listed with its ID number (like the
	// 11 in Game 11: ...) followed by a semicolon-separated list of subsets
	// of cubes that were revealed from the bag (like 3 red, 5 green, 4
	// blue).
	//
	// In game 1, three sets of cubes are revealed from the bag (and then
	// put back again). The first set is 3 blue cubes and 4 red cubes; the
	// second set is 1 red cube, 2 green cubes, and 6 blue cubes; the third
	// set is only 2 green cubes.
	input := "Game 1: 3 blue, 4 red; 1 red, 2 green, 6 blue; 2 green\n" +
		"Game 2: 1 blue, 2 green; 3 green, 4 blue, 1 red; 1 green, 1 blue\n" +
		"Game 3: 8 green, 6 blue, 20 red; 5 blue, 4 red, 13 green; 5 green, 1 red\n" +
		"Game 4: 1 green, 3 red, 6 blue; 3 green, 6 red; 3 green, 15 blue, 14 red\n" +
		"Game 5: 6 red, 1 blue, 3 green; 2 blue, 1 red, 2 green"

	// The Elf would first like to know which games would have been possible
	// if the bag contained only 12 red cubes, 13 green cubes, and 14 blue
	// cubes?
	result, err := part1.GetSecretNumber(strings.NewReader(input), part1.Bag{
		Red:   12,
		Green: 13,
		Blue:  14,
	})
	if err != nil {
		log.Fatalln(err)
	}

	// In the example above, games 1, 2, and 5 would have been possible if
	// the bag had been loaded with that configuration. However, game 3
	// would have been impossible because at one point the Elf showed you 20
	// red cubes at once; similarly, game 4 would also have been impossible
	// because the Elf showed you 15 blue cubes at once. If you add up the
	// IDs of the games that would have been possible, you get 8.
	fmt.Println(result)
	// Output: 8
}
