package part1_test

import (
	"fmt"
	"log"
	"strings"
	"testing"

	"github.com/google/go-cmp/cmp"
	"github.com/google/go-cmp/cmp/cmpopts"

	"source.toby3d.me/toby3d/adventofcode/internal/day5/part1"
)

var filter = cmpopts.IgnoreUnexported(part1.Map{}, part1.Range{})

func TestParseSeeds(t *testing.T) {
	t.Parallel()

	actual, err := part1.ParseSeeds("seeds: 79 14 55 13")
	if err != nil {
		t.Fatal(err)
	}

	expect := []uint64{79, 14, 55, 13}
	if diff := cmp.Diff(actual, expect); diff != "" {
		t.Error(diff)
	}
}

func TestParseMap(t *testing.T) {
	t.Parallel()

	for name, tc := range map[string]struct {
		input  string
		expect *part1.Map
	}{
		"seed2soil": {
			input: "seed-to-soil map:\n" +
				"50 98 2\n" +
				"52 50 48",
			expect: part1.NewMap("seed", "soil",
				part1.NewRange(50, 98, 2),
				part1.NewRange(52, 50, 48),
			),
		},
		"soil2fertilizer": {
			input: "soil-to-fertilizer map:\n" +
				"0 15 37\n" +
				"37 52 2\n" +
				"39 0 15",
			expect: part1.NewMap("soil", "fertilizer",
				part1.NewRange(0, 15, 37),
				part1.NewRange(37, 52, 2),
				part1.NewRange(39, 0, 15),
			),
		},
		"fertilizer2water": {
			input: "fertilizer-to-water map:\n" +
				"49 53 8\n" +
				"0 11 42\n" +
				"42 0 7\n" +
				"57 7 4",
			expect: part1.NewMap("fertilizer", "water",
				part1.NewRange(49, 53, 8),
				part1.NewRange(0, 11, 42),
				part1.NewRange(42, 0, 7),
				part1.NewRange(57, 7, 4),
			),
		},
		"water2light": {
			input: "water-to-light map:\n" +
				"88 18 7\n" +
				"18 25 70",
			expect: part1.NewMap("water", "light",
				part1.NewRange(88, 18, 7),
				part1.NewRange(18, 25, 70),
			),
		},
		"light2temperature": {
			input: "light-to-temperature map:\n" +
				"45 77 23\n" +
				"81 45 19\n" +
				"68 64 13",
			expect: part1.NewMap("light", "temperature",
				part1.NewRange(45, 77, 23),
				part1.NewRange(81, 45, 19),
				part1.NewRange(68, 64, 13),
			),
		},
		"temperature2humidity": {
			input: "temperature-to-humidity map:\n" +
				"0 69 1\n" +
				"1 0 69",
			expect: part1.NewMap("temperature", "humidity",
				part1.NewRange(0, 69, 1),
				part1.NewRange(1, 0, 69),
			),
		},
		"humidity2location": {
			input: "humidity-to-location map:\n" +
				"60 56 37\n" +
				"56 93 4",
			expect: part1.NewMap("humidity", "location",
				part1.NewRange(60, 56, 37),
				part1.NewRange(56, 93, 4),
			),
		},
	} {
		name, tc := name, tc

		t.Run(name, func(t *testing.T) {
			t.Parallel()

			actual, err := part1.ParseMap(strings.NewReader(tc.input))
			if err != nil {
				t.Fatal(err)
			}

			if diff := cmp.Diff(actual, tc.expect, filter); diff != "" {
				t.Error(diff)
			}
		})
	}
}

/*
func TestRange_Map(t *testing.T) {
	t.Parallel()

	input := part1.Rangepart1.NewRange( 50, 98, 2)
	actual := input.Map()
	expect := map[uint64]uint64{
		98: 50,
		99: 51,
	}

	if diff := cmp.Diff(actual, expect, filter); diff != "" {
		t.Error(diff)
	}
}
*/

func TestMaps_Locate(t *testing.T) {
	t.Parallel()

	input := part1.Almanac{Maps: []*part1.Map{
		part1.NewMap("seed", "soil",
			part1.NewRange(50, 98, 2),
			part1.NewRange(52, 50, 48),
		),
		part1.NewMap("soil", "fertilizer",
			part1.NewRange(0, 15, 37),
			part1.NewRange(37, 52, 2),
			part1.NewRange(39, 0, 15),
		),
		part1.NewMap("fertilizer", "water",
			part1.NewRange(49, 53, 8),
			part1.NewRange(0, 11, 42),
			part1.NewRange(42, 0, 7),
			part1.NewRange(57, 7, 4),
		),
		part1.NewMap("water", "light",
			part1.NewRange(88, 18, 7),
			part1.NewRange(18, 25, 70),
		),
		part1.NewMap("light", "temperature",
			part1.NewRange(45, 77, 23),
			part1.NewRange(81, 45, 19),
			part1.NewRange(68, 64, 13),
		),
		part1.NewMap("temperature", "humidity",
			part1.NewRange(0, 69, 1),
			part1.NewRange(1, 0, 69),
		),
		part1.NewMap("humidity", "location",
			part1.NewRange(60, 56, 37),
			part1.NewRange(56, 93, 4),
		),
	}}

	for name, tc := range map[string]struct {
		input  uint64
		expect uint64
	}{
		"1": {79, 82},
		"2": {14, 43},
		"3": {55, 86},
		"4": {13, 35},
	} {
		name, tc := name, tc

		t.Run(name, func(t *testing.T) {
			t.Parallel()

			if actual := input.Locate(tc.input); actual != tc.expect {
				t.Errorf("want %d, got %d", tc.expect, actual)
			}
		})
	}
}

// The almanac (your puzzle input) lists all of the seeds that need to be
// planted. It also lists what type of soil to use with each kind of seed, what
// type of fertilizer to use with each kind of soil, what type of water to use
// with each kind of fertilizer, and so on. Every type of seed, soil, fertilizer
// and so on is identified with a number, but numbers are reused by each
// category - that is, soil 123 and fertilizer 123 aren't necessarily related to
// each other.
func Example() {
	input := "seeds: 79 14 55 13\n" +
		"\n" +
		"seed-to-soil map:\n" +
		"50 98 2\n" +
		"52 50 48\n" +
		"\n" +
		"soil-to-fertilizer map:\n" +
		"0 15 37\n" +
		"37 52 2\n" +
		"39 0 15\n" +
		"\n" +
		"fertilizer-to-water map:\n" +
		"49 53 8\n" +
		"0 11 42\n" +
		"42 0 7\n" +
		"57 7 4\n" +
		"\n" +
		"water-to-light map:\n" +
		"88 18 7\n" +
		"18 25 70\n" +
		"\n" +
		"light-to-temperature map:\n" +
		"45 77 23\n" +
		"81 45 19\n" +
		"68 64 13\n" +
		"\n" +
		"temperature-to-humidity map:\n" +
		"0 69 1\n" +
		"1 0 69\n" +
		"\n" +
		"humidity-to-location map:\n" +
		"60 56 37\n" +
		"56 93 4"

	almanac, err := part1.ParseAlmanac(strings.NewReader(input))
	if err != nil {
		log.Fatalln(err)
	}

	fmt.Println(almanac.FindLowestLocation())
	// Output: 35
}
