package part2

import (
	"bufio"
	"bytes"
	"fmt"
	"io"
	"strconv"
	"strings"
)

type (
	Seed struct {
		Start  uint64
		Length uint64
	}

	Range struct {
		destinationStart uint64
		sourceStart      uint64
		length           uint64
	}

	Map struct {
		source      string
		destination string
		Ranges      []*Range
	}

	Almanac struct {
		Seeds []Seed
		Maps  []*Map
	}
)

func NewRange(dst, src, l uint64) *Range {
	r := &Range{
		destinationStart: dst,
		sourceStart:      src,
		length:           l,
	}

	return r
}

func (r Range) Contains(source uint64) bool {
	return r.sourceStart <= source &&
		source < r.sourceStart+r.length
}

func (r Range) Locate(source uint64) uint64 {
	return r.destinationStart + (source - r.sourceStart)
}

func NewMap(src, dst string, ranges ...*Range) *Map {
	m := &Map{
		source:      src,
		destination: dst,
		Ranges:      ranges,
	}

	return m
}

func (m Map) Locate(source uint64) uint64 {
	for i := range m.Ranges {
		if !m.Ranges[i].Contains(source) {
			continue
		}

		source = m.Ranges[i].Locate(source)

		break
	}

	return source
}

func (a Almanac) Locate(seed uint64) uint64 {
	loc := seed

	for i := range a.Maps {
		loc = a.Maps[i].Locate(loc)
	}

	return loc
}

func ParseSeeds(str string) ([]Seed, error) {
	result := make([]Seed, 0)
	data := strings.Fields(str[7:])

	for i := 0; i < len(data); i += 2 {
		seed := new(Seed)
		// result := append(result, Seed{start: data[i], length: data[i]})

		for src, dst := range map[string]*uint64{
			data[i]:   &seed.Start,
			data[i+1]: &seed.Length,
		} {
			var err error
			if *dst, err = strconv.ParseUint(src, 10, 64); err != nil {
				return nil, fmt.Errorf("cannot parse seed value: %w", err)
			}
		}

		result = append(result, *seed)
	}

	return result, nil
}

func ParseMap(r io.Reader) (*Map, error) {
	scanner := bufio.NewScanner(r)
	scanner.Split(bufio.ScanLines)

	var source, destination string

	if scanner.Scan() {
		source, destination = parseMapDirection(scanner.Text())
	}

	ranges := make([]*Range, 0)
	for scanner.Scan() {
		parts := strings.Split(scanner.Text(), " ")

		var dst, src, l uint64

		for i, dst := range []*uint64{&dst, &src, &l} {
			var err error
			if *dst, err = strconv.ParseUint(parts[i], 10, 64); err != nil {
				return nil, fmt.Errorf("cannot parse map range part '%s' as number: %w", parts[i], err)
			}
		}

		ranges = append(ranges, NewRange(dst, src, l))
	}

	return NewMap(source, destination, ranges...), nil
}

func ParseAlmanac(r io.Reader) (*Almanac, error) {
	scanner := bufio.NewScanner(r)
	scanner.Split(bufio.ScanLines)

	result := &Almanac{
		Seeds: make([]Seed, 0),
		Maps:  nil,
	}

	if scanner.Scan() {
		var err error
		if result.Seeds, err = ParseSeeds(scanner.Text()); err != nil {
			return nil, fmt.Errorf("cannot parse almanac seeds: %w", err)
		}
	}

	maps := make([][]byte, 0)
	currentMap := make([]byte, 0)

	for scanner.Scan() {
		line := scanner.Bytes()

		if len(line) != 0 {
			currentMap = append(currentMap, line...)
			currentMap = append(currentMap, "\n"...)

			continue
		}

		if len(currentMap) == 0 {
			continue
		}

		maps = append(maps, currentMap)
		currentMap = make([]byte, 0)
	}

	if len(currentMap) != 0 {
		maps = append(maps, currentMap)
	}

	result.Maps = make([]*Map, len(maps))

	for i := range maps {
		var err error
		if result.Maps[i], err = ParseMap(bytes.NewReader(maps[i])); err != nil {
			return nil, fmt.Errorf("cannot parse almanac map: %w", err)
		}
	}

	return result, nil
}

func (a Almanac) FindLowestLocation() uint64 {
	result := a.Seeds[0].Start

	for i := range a.Seeds {
		for seed := a.Seeds[i].Start; seed < a.Seeds[i].Start+a.Seeds[i].Length; seed++ {
			result = min(result, a.Locate(seed))
		}
	}

	return result
}

func parseMapDirection(line string) (string, string) {
	index := strings.IndexByte(line, '-')
	src := line[:index]
	index += 4
	dst := line[index:strings.IndexByte(line, ' ')]

	return src, dst
}
