package part1

import (
	"bufio"
	"fmt"
	"io"
	"sort"
	"strconv"
	"strings"
)

type (
	Card struct{ power uint8 }

	Hand struct {
		cards    [5]Card
		handType HandType
		bid      uint64
	}

	HandType struct{ power uint8 }

	CamelCards []Hand
)

var (
	CardUnd Card = Card{0}  // 0
	CardA   Card = Card{13} // 13
	CardK   Card = Card{12} // 12
	CardQ   Card = Card{11} // 11
	CardJ   Card = Card{10} // 10
	CardT   Card = Card{9}  // 9
	Card9   Card = Card{8}  // 8
	Card8   Card = Card{7}  // 7
	Card7   Card = Card{6}  // 6
	Card6   Card = Card{5}  // 5
	Card5   Card = Card{4}  // 4
	Card4   Card = Card{3}  // 3
	Card3   Card = Card{2}  // 2
	Card2   Card = Card{1}  // 1
)

var (
	HandTypeUnd           HandType = HandType{0} // 0
	HandTypeFiveOfTheKind HandType = HandType{7} // 7
	HandTypeFourOfAKind   HandType = HandType{6} // 6
	HandTypeFullHouse     HandType = HandType{5} // 5
	HandTypeThreeOfAKind  HandType = HandType{4} // 4
	HandTypeTwoPair       HandType = HandType{3} // 3
	HandTypeOnePair       HandType = HandType{2} // 2
	HandTypeHighCard      HandType = HandType{1} // 1
)

var bytesToCards map[byte]Card = map[byte]Card{
	'2': Card2,
	'3': Card3,
	'4': Card4,
	'5': Card5,
	'6': Card6,
	'7': Card7,
	'8': Card8,
	'9': Card9,
	'A': CardA,
	'J': CardJ,
	'K': CardK,
	'Q': CardQ,
	'T': CardT,
}

func ParseCard(b byte) (Card, error) {
	if card, ok := bytesToCards[b]; ok {
		return card, nil
	}

	return CardUnd, fmt.Errorf("cannot parse '%s' as Card", string(b))
}

func NewHand(cards [5]Card) Hand {
	result := Hand{
		cards:    cards,
		handType: HandTypeUnd,
		bid:      0,
	}
	kinds := make(map[Card]uint8)

	for i := range cards {
		if cards[i] == CardUnd {
			continue
		}

		if _, ok := kinds[cards[i]]; !ok {
			kinds[cards[i]] = 0
		}

		kinds[cards[i]]++
	}

	switch len(kinds) {
	case 5:
		result.handType = HandTypeHighCard
	case 4:
		for _, count := range kinds {
			if count != 2 {
				continue
			}

			result.handType = HandTypeOnePair

			break
		}
	case 3:
		for _, count := range kinds {
			switch count {
			default:
				continue
			case 3:
				result.handType = HandTypeThreeOfAKind
			case 2:
				result.handType = HandTypeTwoPair
			}

			break
		}
	case 2:
		for _, count := range kinds {
			switch count {
			default:
				continue
			case 4:
				result.handType = HandTypeFourOfAKind
			case 2, 3:
				result.handType = HandTypeFullHouse
			}

			break
		}
	case 1:
		result.handType = HandTypeFiveOfTheKind
	}

	return result
}

func (h Hand) Type() HandType {
	return h.handType
}

func (h Hand) IsWinAgainst(player Hand) bool {
	if h.handType.power > player.handType.power {
		return true
	}

	for i := 0; i < 5; i++ {
		if h.cards[i].power == player.cards[i].power {
			continue
		}

		return h.cards[i].power > player.cards[i].power
	}

	return false
}

func ParseCamelCards(r io.Reader) (CamelCards, error) {
	scanner := bufio.NewScanner(r)
	scanner.Split(bufio.ScanLines)

	result := make(CamelCards, 0)

	for scanner.Scan() {
		line := scanner.Text()
		index := strings.IndexRune(line, ' ')

		var cards [5]Card
		for i := range line[:index] {
			var err error
			if cards[i], err = ParseCard(line[i]); err != nil {
				return nil, fmt.Errorf("cannot parse card in hand '%s': %w", line, err)
			}
		}

		bid, err := strconv.ParseUint(line[index+1:], 10, 16)
		if err != nil {
			return nil, fmt.Errorf("cannot parse hand '%s' bid as number: %w", line, err)
		}

		hand := NewHand(cards)
		hand.bid = bid
		result = append(result, hand)
	}

	return result, nil
}

func (cc CamelCards) CountTotalWinnings() uint64 {
	// NOTE(toby3d): sort by rank
	sort.Slice(cc, func(i, j int) bool {
		return cc[i].handType.power < cc[j].handType.power ||
			!cc[i].IsWinAgainst(cc[j])
	})

	var result uint64
	for i := range cc {
		result += cc[i].bid * (uint64(i) + 1)
	}

	return result
}

func (c Card) String() string {
	for b, card := range bytesToCards {
		if card != c {
			continue
		}

		return string(b)
	}

	return "und"
}

func (ht HandType) GoString() string {
	return "part1.HandType(" + strconv.FormatUint(uint64(ht.power), 10) + ")"
}
